#include "../../sf.h"

#include <sf/diag/log.h>

#include "../Application.h"
#include "AndroidApplication.h"
#include "AudioManager.h"
#include "jni.h"
#include "../application.h"
#include "../TextureManager.h"
#include "../../gui/widgetmanager.h"
#include "../../graphics/renderer.h"

#include "../../gui/gestures/GestureRecognizers.h"
#include "../../gui/gestures/DragGestureRecognizer.h"
#include "../../gui/gestures/SwipeGestureRecognizer.h"
#include "../../gui/gestures/LongPressGestureRecognizer.h"
#include "../../gui/gestures/RotateGestureRecognizer.h"

namespace sf
{
namespace core
{

static CAndroidApplication *g_Me = 0;
static jmethodID g_GameThreadRender = 0;
static jmethodID g_GameThreadGetScreenSize = 0;
static jmethodID g_GameThreadHandleEvents = 0;
static jmethodID g_GameThreadShowLink = 0;
static jmethodID g_GameThreadStopGame = 0;

AndroidTouchSubscribeContext* CAndroidApplication::m_TouchesCallbackContext = 0;

CAndroidApplication::CAndroidApplication()
:   m_EndModal(false),
    m_Stop(false),
    m_LogicalWidth(0),
    m_LogicalHeight(0),
    m_ScreenWidth(0),
    m_ScreenHeight(0),
    m_ModalDepth(0),
    m_Paused(false),
    m_SystemGesturePinch(false),
    m_SystemGestureTap(false),
    m_SystemGestureDoubleTap(false)
{
    SFASSERT(!g_Me);
    g_Me = this;

    memset(m_ButtonCallbacks, 0, ButtonCallbackCount*sizeof(m_ButtonCallbacks[0]));

    SFASSERT(g_JavaVM);
    
    jclass game_thread_class = JNIGetClass("com/stargaze/sf/GameThread");
    
    g_GameThreadRender = JNIGetObjectMethod(game_thread_class, "render", "()V");
    g_GameThreadGetScreenSize = JNIGetObjectMethod(game_thread_class, "getScreenSize", "()[I");
    g_GameThreadHandleEvents = JNIGetObjectMethod(game_thread_class, "handleEvents", "()V");
    g_GameThreadShowLink = JNIGetObjectMethod(game_thread_class, "showLink", "(Ljava/lang/String;)V");
	g_GameThreadStopGame = sf::core::JNIGetObjectMethod(game_thread_class, "stopGame", "()V");

    
    jintArray jarray = (jintArray) g_JavaEnv->CallObjectMethod(g_GameThread, g_GameThreadGetScreenSize);
    jint* size = g_JavaEnv->GetIntArrayElements(jarray, 0);
    if (size)
    {
    	SetRenderRect(size[0], size[1]);
        g_JavaEnv->ReleaseIntArrayElements(jarray, size, 0);
    }
    g_JavaEnv->DeleteLocalRef(jarray);
}

CAndroidApplication::~CAndroidApplication()
{
}

bool CAndroidApplication::Init()
{
    sf::core::g_Application->EnableSystemGestureRecognizer(sf::core::CApplication::GesturePinchRecognizer, true);
    sf::core::g_Application->EnableSystemGestureRecognizer(sf::core::CApplication::GestureTapRecognizer, true);
    sf::core::g_Application->EnableSystemGestureRecognizer(sf::core::CApplication::GestureTapDoubleRecognizer, true);
    
    // currently unsupported by system gestures must be emulated by sf
    
    sf::core::g_Application->EnableSystemGestureRecognizer(sf::core::CApplication::GesturePanRecognizer, false);
    sf::core::g_Application->EnableSystemGestureRecognizer(sf::core::CApplication::GestureRotationRecognizer, false);
    sf::core::g_Application->EnableSystemGestureRecognizer(sf::core::CApplication::GestureSwipeRecognizer, false);
    sf::core::g_Application->EnableSystemGestureRecognizer(sf::core::CApplication::GestureLongPressRecognizer, false);
        
    sf::core::g_Application->GetWidgetManager().GestureRecognizers()->AddGestureRecognizer(sf::gui::gestures::CBaseGestureRecognizer::SharedPtr(new sf::gui::gestures::CDragGestureRecognizer));
    sf::core::g_Application->GetWidgetManager().GestureRecognizers()->AddGestureRecognizer(sf::gui::gestures::CBaseGestureRecognizer::SharedPtr(new sf::gui::gestures::CSwipeGestureRecognizer));
    sf::core::g_Application->GetWidgetManager().GestureRecognizers()->AddGestureRecognizer(sf::gui::gestures::CBaseGestureRecognizer::SharedPtr(new sf::gui::gestures::CLongPressGestureRecognizer));
    sf::core::g_Application->GetWidgetManager().GestureRecognizers()->AddGestureRecognizer(sf::gui::gestures::CBaseGestureRecognizer::SharedPtr(new sf::gui::gestures::CRotateGestureRecognizer));
    
    return true;
}

void CAndroidApplication::Run()
{
    MessageCycle();
}

void CAndroidApplication::Stop()
{
	if (m_Stop) return;

	//closing audio threads
	sf::core::g_AudioManager::Instance().StopAll(0);
	sf::core::g_AudioManager::Instance().Update();
	//
	sf::core::g_JavaEnv->CallVoidMethod(sf::core::g_GameThread, g_GameThreadStopGame);
    m_Stop = true;
}

void CAndroidApplication::Shutdown()
{
}

void CAndroidApplication::Pause()
{
	if (m_Paused) return;
	m_Paused = true;

	g_TimeManager::Instance().Pause(true);
	g_AudioManager::Instance().PauseAll(true);
	if (g_Application)
		g_Application->SaveAppState();

	g_TextureManager::Instance().InvalidateSurfaces();
}

void CAndroidApplication::Resume()
{
	if (!m_Paused) return;
	m_Paused = false;

	g_TimeManager::Instance().Pause(false);
	g_AudioManager::Instance().PauseAll(false);
}

void CAndroidApplication::SetButtonCallback(ButtonCallbackTypes _type, ButtonCallback* _func)
{
	m_ButtonCallbacks[_type] = _func;
}

void CAndroidApplication::OnMenuButton()
{
	// g_Application->GetWidgetManager().DispatchMenuButtonPressed();
	if (m_ButtonCallbacks[ButtonCallbackMenu]) return (*m_ButtonCallbacks[ButtonCallbackMenu])();
}

void CAndroidApplication::OnBackButton()
{
	 g_Application->GetWidgetManager().DispatchBackButtonPressed();
	if (m_ButtonCallbacks[ButtonCallbackBack]) 
		return (*m_ButtonCallbacks[ButtonCallbackBack])();
}

void CAndroidApplication::OnSearchButton()
{
	if (m_ButtonCallbacks[ButtonCallbackSearch]) return (*m_ButtonCallbacks[ButtonCallbackSearch])();
}

void CAndroidApplication::ShowLink(const char* _link)
{
    jstring path = g_JavaEnv->NewStringUTF(_link);

    g_JavaEnv->CallVoidMethod(g_GameThread, g_GameThreadShowLink, path);

    g_JavaEnv->DeleteLocalRef(path);
}

void CAndroidApplication::DoModal()
{
    MessageCycle();
}

void CAndroidApplication::StopModal()
{
    m_EndModal = true;
}

void CAndroidApplication::SetRenderRect(int _width, int _height)
{
    if (_width > _height)
    {
	   	m_LogicalWidth = _width;
       	m_LogicalHeight = _height;
    }
    else
    {
       	m_LogicalWidth = _height;
       	m_LogicalHeight = _width;
    }
    m_ScreenWidth = m_LogicalWidth;
    m_ScreenHeight = m_LogicalHeight;
}

bool CAndroidApplication::ChangeScreenMode(bool _fullscreen, bool _32bpp, size_t _width, size_t _height) 
{
    return true;
}

bool CAndroidApplication::GetOriginalDesktopDimentions(size_t &_width, size_t &_height)
{
    return true;
}

static inline void AdaptPos(float _x, float _y, float& _out_x, float& _out_y)
{
	const sf::misc::FloatVector& scale = g_Application->GetRenderer()->GetScale();
	const sf::misc::FloatVector& offset = g_Application->GetRenderer()->GetAspectOffset();
	_out_x = (_x - offset.X)/scale.X;
	_out_y = (_y - offset.Y)/scale.Y;
}

void CAndroidApplication::OnTouchBegin(const gui::Touches& _touches, int _timestamp)
{
	g_Application->GetWidgetManager().DispatchTouchBegin(_touches, _timestamp);
}

void CAndroidApplication::OnTouchMove(const gui::Touches& _touches, int _timestamp)
{
	g_Application->GetWidgetManager().DispatchTouchMove(_touches, _timestamp);
}

void CAndroidApplication::OnTouchEnd(const gui::Touches& _touches, int _timestamp)
{
	g_Application->GetWidgetManager().DispatchTouchEnd(_touches, _timestamp);
}

void CAndroidApplication::AddScrollCallback(void * _arg, AndroidScrollCallback * _callback)
{
	m_ScrollCallbacks[_callback] = _arg;
}


void CAndroidApplication::RemoveScrollCallback(void * _arg, AndroidScrollCallback * _callback)
{
	m_ScrollCallbacks.erase(_callback);
}

void CAndroidApplication::OnScroll(int _pid, float _x, float _y, float _dx, float _dy) {
	for(ScrollCallbacksT::iterator it = m_ScrollCallbacks.begin(), end = m_ScrollCallbacks.end(); it != end; it++)
		it->first(it->second, _x, _y, _dx, _dy);
}

void CAndroidApplication::OnPinch(float x, float y, float _scale, float _velocity, int _state)
{
    if (m_SystemGesturePinch)
    {
        sf::core::g_Application->GetWidgetManager().DispatchPinchGesture(sf::misc::IntVector(x, y)
            , _scale, _velocity, sf::gui::CBaseWidget::GestureStates(_state));
    }
}

void CAndroidApplication::OnTap(float x, float y) {
    if (m_SystemGestureTap)
    {
        sf::core::g_Application->GetWidgetManager().DispatchTapGesture(sf::misc::IntVector(x, y));
    }
}

void CAndroidApplication::OnDoubleTap(float x, float y) {
    if (m_SystemGestureDoubleTap)
    {
        sf::core::g_Application->GetWidgetManager().DispatchDoubleTapGesture(sf::misc::IntVector(x, y));
    }
}

EventInformation& CAndroidApplication::GetCurrentEvent()
{
    static EventInformation e;
    return e;
}

void CAndroidApplication::DefaultWindowProc() {}

void CAndroidApplication::GetWindowClientRect(misc::IntRect &_rc)
{
    _rc.X = _rc.Y = 0;
    _rc.Width = m_LogicalWidth;
    _rc.Height = m_LogicalHeight;
}

void CAndroidApplication::AdjustClientRectToWindow(misc::IntRect &_rc) {}

void CAndroidApplication::GetDesktopWindowedSpace(misc::IntRect &_rc)
{
    _rc.X = _rc.Y = 0;
    _rc.Width = m_LogicalWidth;
    _rc.Height = m_LogicalHeight;
}

void CAndroidApplication::ScreenCoordsIntoClient(misc::IntVector& _pos) {}

void CAndroidApplication::ClientCoordsIntoScreen(misc::IntVector& _pos) {}

void CAndroidApplication::SetMouseCursorPos(const misc::IntVector& _pos) {}

void CAndroidApplication::GetMouseCursorPos(misc::IntVector& _pos) {}

void CAndroidApplication::SetSysCursor(gui::SysCursor _cursor, bool _show_now) {}

gui::SysCursor CAndroidApplication::GetSysCursor()
{ 
    return gui::SysCursor(); 
}

void CAndroidApplication::ShowSysCursor(bool _show) {}

bool CAndroidApplication::IsSysCursorShown() 
{
    return false;
}

void CAndroidApplication::PublicAppDraw()
{
    AppDraw();
}

CAndroidApplication* CAndroidApplication::GetMe()
{
    return g_Me;
}

void CAndroidApplication::MessageCycle()
{
    ++m_ModalDepth;

    while (!m_Stop && !m_EndModal)
    {
    	// GameThread.handleEvents
    	g_JavaEnv->CallVoidMethod(g_GameThread, g_GameThreadHandleEvents);

    	if (m_Stop || m_EndModal) break;
    	if (m_Paused) continue;

        AppUpdate();
        // GameThread.render()
        g_JavaEnv->CallVoidMethod(g_GameThread, g_GameThreadRender);
    }

    m_EndModal = false;

   --m_ModalDepth;
}
	
void CAndroidApplication::EnableSystemGestureRecognizer(int _recognizerType, bool _enable)
{
    switch(_recognizerType)
    {
    case sf::core::CApplication::GesturePinchRecognizer:
        m_SystemGesturePinch = _enable;
        break;
            
    case sf::core::CApplication::GestureTapRecognizer:
        m_SystemGestureTap = _enable;
        break;
        
    case sf::core::CApplication::GestureTapDoubleRecognizer:
        m_SystemGestureDoubleTap = _enable;
        break;
        
    default:
        break;
    }
}


} // core
} // sf

extern "C" void Java_com_stargaze_sf_CAndroidApplication_appDraw(JNIEnv* _env, jobject _this)
{
    SFASSERT(sf::core::g_Me);    	
   	sf::core::g_Me->PublicAppDraw();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onTouchBegin(JNIEnv* _env, jobject _this, jint _count, jintArray _x, jintArray _y, jint _timestamp)
{
    SFASSERT(sf::core::g_Me);
    jboolean isCopy = JNI_FALSE;
    jint *jarrx = _env->GetIntArrayElements(_x, &isCopy);
    jint *jarry = _env->GetIntArrayElements(_y, &isCopy);

    sf::gui::Touches touchesContainer;
    for (int i = 0; i < _count; i++)
    {
    	touchesContainer.push_back(sf::misc::IntVector(jarrx[i], jarry[i]));
    }
   	sf::core::g_Me->OnTouchBegin(touchesContainer, _timestamp);

   	_env->ReleaseIntArrayElements(_x, jarrx, JNI_ABORT);
   	_env->ReleaseIntArrayElements(_y, jarry, JNI_ABORT);
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onTouchMove(JNIEnv* _env, jobject _this, jint _count, jintArray _x, jintArray _y, jint _timestamp)
{
	SFASSERT(sf::core::g_Me);
	jboolean isCopy = JNI_FALSE;
	jint *jarrx = _env->GetIntArrayElements(_x, &isCopy);
	jint *jarry = _env->GetIntArrayElements(_y, &isCopy);

	sf::gui::Touches touchesContainer;
	for (int i = 0; i < _count; i++) {
		touchesContainer.push_back(sf::misc::IntVector(jarrx[i], jarry[i]));
	}
	sf::core::g_Me->OnTouchMove(touchesContainer, _timestamp);

	_env->ReleaseIntArrayElements(_x, jarrx, JNI_ABORT);
	_env->ReleaseIntArrayElements(_y, jarry, JNI_ABORT);
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onTouchEnd(JNIEnv* _env, jobject _this, jint _count, jintArray _x, jintArray _y, jint _timestamp)
{
    SFASSERT(sf::core::g_Me);
	jboolean isCopy = JNI_FALSE;
	jint *jarrx = _env->GetIntArrayElements(_x, &isCopy);
	jint *jarry = _env->GetIntArrayElements(_y, &isCopy);

	sf::gui::Touches touchesContainer;
	for (int i = 0; i < _count; i++) {
		touchesContainer.push_back(sf::misc::IntVector(jarrx[i], jarry[i]));
	}
	sf::core::g_Me->OnTouchEnd(touchesContainer, _timestamp);

	_env->ReleaseIntArrayElements(_x, jarrx, JNI_ABORT);
	_env->ReleaseIntArrayElements(_y, jarry, JNI_ABORT);
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onScroll(JNIEnv* _env,
		jobject _this, jint _pid, jfloat _x, jfloat _y, jfloat _dx, jfloat _dy) {
	try {
		SFASSERT(sf::core::g_Me);
		sf::core::g_Me->OnScroll(_pid, _x, _y, _dx, _dy);

	} catch (std::exception & ex) {
		SFLOGA("CAndroidApplication", sf::diag::LogError,
				"in onScroll exception: %s", ex.what());
	} catch (...) {
		SFLOGA("CAndroidApplication", sf::diag::LogError,
				"in onScroll unknown exception");
	}
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_stop(JNIEnv* _env, jobject _this)
{
	SFASSERT(sf::core::g_Me);
	sf::core::g_Me->Stop();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_pause(JNIEnv* _env, jobject _this)
{
	if (sf::core::g_Me)
		sf::core::g_Me->Pause();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_resume(JNIEnv* _env, jobject _this)
{
	if (sf::core::g_Me)
		sf::core::g_Me->Resume();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onMenuButton(JNIEnv* _env, jobject _this)
{
	if (sf::core::g_Me)
		sf::core::g_Me->OnMenuButton();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onBackButton(JNIEnv* _env, jobject _this)
{
	if (sf::core::g_Me)
		sf::core::g_Me->OnBackButton();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onSearchButton(JNIEnv* _env, jobject _this)
{
	if (sf::core::g_Me)
		sf::core::g_Me->OnSearchButton();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onRenderRectChanged(JNIEnv* _env, jobject _this, jint _width, jint _height)
{
	if(sf::core::g_Me)
		sf::core::g_Me->SetRenderRect(_width, _height);

	if (sf::core::g_Application) {
		sf::graphics::Resolution res;
		res.Width = _width;
		res.Height = _height;
		sf::graphics::CRenderer::ResolutionsT resolutions;
		resolutions.push_back(res);
		sf::core::g_Application->GetRenderer()->SetAvailableResolutions(resolutions);
		sf::core::g_Application->ChangeScreenMode(false, false, _width, _height);
	}
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onPinch(JNIEnv* _env, jobject _this, jfloat x, jfloat y, jfloat _scale, jfloat _velocity, jint _state)
{
	if(sf::core::g_Me)
		sf::core::g_Me->OnPinch(x, y, _scale, _velocity, _state);
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onTap(JNIEnv* _env, jobject _this, jfloat x, jfloat y)
{
	if(sf::core::g_Me)
		sf::core::g_Me->OnTap(x, y);
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onDoubleTap(JNIEnv* _env, jobject _this, jfloat x, jfloat y)
{
	if(sf::core::g_Me)
		sf::core::g_Me->OnDoubleTap(x, y);
}

extern "C" bool Java_com_stargaze_sf_CAndroidApplication_isDebugAssembly(JNIEnv* _env, jobject _this)
{
	#ifdef SF_DEVELOPMENT
	return true;
	#else
	return false;
	#endif
}


